try:
    # requests uses simplejson if its present, so we need to do the same.
    # see https://github.com/psf/requests/issues/4842
    import simplejson as json
except ImportError:
    import json
import chess
from .types import *

BACK_RANKS = list(chess.SquareSet(chess.BB_BACKRANKS))


def add_pawn_queen_promotion(board: chess.Board, move: chess.Move) -> chess.Move:
    piece = board.piece_at(move.from_square)
    if piece is not None and piece.piece_type == chess.PAWN and move.to_square in BACK_RANKS and move.promotion is None:
        move = chess.Move(move.from_square, move.to_square, chess.QUEEN)
    return move


def is_psuedo_legal_castle(board: chess.Board, move: chess.Move):
    return board.is_castling(move) and not is_illegal_castle(board, move)


def is_illegal_castle(board: chess.Board, move: chess.Move) -> bool:
    if not board.is_castling(move):
        return False

    # illegal without kingside rights
    if board.is_kingside_castling(move) and not board.has_kingside_castling_rights(board.turn):
        return True

    # illegal without queenside rights
    if board.is_queenside_castling(move) and not board.has_queenside_castling_rights(board.turn):
        return True

    # illegal if any pieces are between king & rook
    rook_square = chess.square(7 if board.is_kingside_castling(move) else 0, chess.square_rank(move.from_square))
    between_squares = chess.SquareSet(chess.BB_BETWEEN[move.from_square][rook_square])
    if any(map(lambda s: board.piece_at(s), between_squares)):
        return True

    # its legal
    return False


def slide_move(board: chess.Board, move: chess.Move) -> Optional[chess.Move]:
    psuedo_legal_moves = list(board.generate_pseudo_legal_moves())
    squares = list(chess.SquareSet(chess.BB_BETWEEN[move.from_square][move.to_square])) + [move.to_square]
    squares = sorted(squares, key=lambda s: chess.square_distance(s, move.from_square), reverse=True)
    for slide_square in squares:
        revised = chess.Move(move.from_square, slide_square, move.promotion)
        if revised in psuedo_legal_moves:
            return revised
    return None


def capture_square_of_move(board: chess.Board, move: Optional[chess.Move]) -> Optional[Square]:
    capture_square = None
    if move is not None and board.is_capture(move):
        if board.is_en_passant(move):
            # taken from :func:`chess.Board.push()`
            down = -8 if board.turn == chess.WHITE else 8
            capture_square = board.ep_square + down
        else:
            capture_square = move.to_square
    return capture_square


def without_opponent_pieces(board: chess.Board) -> chess.Board:
    """Returns a copy of `board` with the opponent's pieces removed."""
    b = board.copy()
    for piece_type in chess.PIECE_TYPES:
        for sq in b.pieces(piece_type, not board.turn):
            b.remove_piece_at(sq)
    return b


def moves_without_opponent_pieces(board: chess.Board) -> List[chess.Move]:
    """Generates moves on `board` with the opponent's pieces removed."""
    return list(without_opponent_pieces(board).generate_pseudo_legal_moves())


def pawn_capture_moves_on(board: chess.Board) -> List[chess.Move]:
    """Generates all pawn captures on `board`, even if there is no piece to capture. All promotion moves are included."""
    pawn_capture_moves = []

    no_opponents_board = without_opponent_pieces(board)

    for pawn_square in board.pieces(chess.PAWN, board.turn):
        for attacked_square in board.attacks(pawn_square):
            # skip this square if one of our own pieces are on the square
            if no_opponents_board.piece_at(attacked_square):
                continue

            pawn_capture_moves.append(chess.Move(pawn_square, attacked_square))

            # add in promotion moves
            if attacked_square in chess.SquareSet(chess.BB_BACKRANKS):
                for piece_type in chess.PIECE_TYPES[1:-1]:
                    pawn_capture_moves.append(chess.Move(pawn_square, attacked_square, promotion=piece_type))

    return pawn_capture_moves


class ChessJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, chess.Piece):
            return {
                'type': 'Piece',
                'value': o.symbol(),
            }
        elif isinstance(o, chess.Move):
            return {
                'type': 'Move',
                'value': o.uci(),
            }
        elif isinstance(o, chess.Board):
            return {
                'type': 'Board',
                'value': o.fen(),
            }
        elif isinstance(o, WinReason):
            return {
                'type': 'WinReason',
                'value': o.name,
            }
        return super().default(o)


class ChessJSONDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        hook = self._object_hook
        if 'object_hook' in kwargs:
            original_hook = kwargs.pop('object_hook')
            hook = lambda obj: self._object_hook(original_hook(obj))
        super().__init__(object_hook=hook, *args, **kwargs)

    def _object_hook(self, obj):
        if 'type' in obj:
            if obj['type'] == 'Piece':
                return chess.Piece.from_symbol(obj['value'])
            elif obj['type'] == 'Move':
                return chess.Move.from_uci(obj['value'])
            elif obj['type'] == 'Board':
                return chess.Board(fen=obj['value'])
            elif obj['type'] == 'WinReason':
                return WinReason[obj['value']]
        return obj



import argparse
import requests
import multiprocessing
import time
import getpass
import traceback
from datetime import datetime
from reconchess import load_player, play_remote_game


class RBCServer:
    def __init__(self, server_url, auth):
        self.invitations_url = '{}/api/invitations'.format(server_url)
        self.user_url = '{}/api/users'.format(server_url)
        self.me_url = '{}/api/users/me'.format(server_url)
        self.session = requests.Session()
        self.session.auth = auth

    def is_connected(self):
        try:
            response = self.session.post(self.me_url)
            if response.status_code == 401:
                print('Authentication Error!')
                print(response.text)
                quit()
            return response.status_code == 200 and response.json()['username'] == self.session.auth[0]
        except requests.RequestException:
            pass
        return False

    def set_max_games(self, max_games):
        self.session.post('{}/max_games'.format(self.me_url), json={'max_games': max_games})

    def get_active_users(self):
        response = self.session.get('{}/'.format(self.user_url))
        return response.json()['usernames']

    def send_invitation(self, opponent, color):
        response = self.session.post('{}/'.format(self.invitations_url), json={
            'opponent': opponent,
            'color': color,
        })
        return response.json()['game_id']

    def get_invitations(self):
        response = self.session.get('{}/'.format(self.invitations_url))
        return response.json()['invitations']

    def accept_invitation(self, invitation_id):
        response = self.session.post('{}/{}'.format(self.invitations_url, invitation_id))
        return response.json()['game_id']


def accept_invitation_and_play(server_url, auth, invitation_id, bot_cls):
    print('[{}] Accepting invitation {}.'.format(datetime.now(), invitation_id))

    server = RBCServer(server_url, auth)
    game_id = server.accept_invitation(invitation_id)

    print('[{}] Invitation {} accepted. Playing game {}.'.format(datetime.now(), invitation_id, game_id))

    try:
        play_remote_game(server_url, game_id, auth, bot_cls())
        print('[{}] Finished game {}'.format(datetime.now(), game_id))
    except:
        print('[{}] Fatal error in game {}:'.format(datetime.now(), game_id))
        traceback.print_exc()


def listen_for_invitations(server_url, auth, bot_cls, max_concurrent_games):
    server = RBCServer(server_url, auth)

    connected = False
    queued_invitations = set()
    with multiprocessing.Pool(processes=max_concurrent_games) as pool:
        while True:
            while not server.is_connected():
                connected = False
                print('[{}] Could not connect to server... waiting 60 seconds before trying again.'.format(
                    datetime.now()))
                time.sleep(60)

            if not connected:
                print('[{}] Connected successfully to server!'.format(datetime.now()))
                connected = True
                server.set_max_games(max_concurrent_games)

            try:
                invitations = server.get_invitations()
                unqueued_invitations = set(invitations) - queued_invitations
                for invitation_id in unqueued_invitations:
                    print('[{}] Received invitation {}.'.format(datetime.now(), invitation_id))
                    pool.apply_async(accept_invitation_and_play, args=(server_url, auth, invitation_id, bot_cls))
                    queued_invitations.add(invitation_id)
            except requests.RequestException as e:
                print(e)

            time.sleep(5)


def ask_for_username():
    return input('Username: ')


def ask_for_password():
    return getpass.getpass()


def ask_for_auth():
    return ask_for_username(), ask_for_password()


def main():
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('bot_path', help='Path to bot source or bot module name.')
    parser.add_argument('--username', default=None, help='Username for login. Enter with prompt if not specified.')
    parser.add_argument('--password', default=None, help='Password for login. Enter with prompt if not specified.')
    parser.add_argument('--server-url', default='https://rbc.jhuapl.edu', help='URL of the server.')
    parser.add_argument('--max-concurrent-games', type=int, default=4,
                        help='The maximum number of games to play at the same time.')
    args = parser.parse_args()

    bot_name, bot_cls = load_player(args.bot_path)

    username = ask_for_username() if args.username is None else args.username
    password = ask_for_password() if args.password is None else args.password
    auth = username, password

    listen_for_invitations(args.server_url, auth, bot_cls, args.max_concurrent_games)


if __name__ == '__main__':
    main()
